; EBOOT.ASM                             August 2nd, 2001 - 21:23    Sean Pringle
; Copyright (c) 2001 Sean Pringle 

; Assembler code written for NASM 0.98
; Boot Sector code for loading and jumping to execute the ENTH Forth kernel.

[BITS 16]
[ORG 07C00h]
jmp start
dot    db '.',0
query  db '?',0
greet  db 'Loading ENTH',0
vesa   db 13,10,'Video',0
head   db 1                             ; head number
track  db 0                             ; track number
offset dw 2400h                         ; offset into kernel load segment
count  db 0                             ; disk read count
limit  db 7                             ; disk read limit
retry  db 0                             ; disk read retry limit
error  db 0                             ; error flag
reads  db 5
dest   dw 1000h

boot   db 0                             ; boot drive

; ******************************************************************************

start:                                  ; untangle the bootlaces!
        xor ax, ax                      ; make sure we have the right data seg
        mov ds, ax
        cli
        mov ax, 0x5000
        mov ss, ax
        mov sp, 0xFFFF
        sti

readdisk:                               ; read system kernel from the boot disk
        mov [boot], dl                  ; save boot drive
        mov ax, 0                       ; al = 00h reset diskette/hard drive
        int 13h                         ; reset disk controller
        jnc sayhi
        mov si, query
        call text
        jmp $

sayhi:                                  ; yes, we are alive
        mov si, greet
        call text

getkernel:
        mov ax, [dest]
        mov es, ax
        call readimage
        call readcheck
        mov byte [head], 1
        mov byte [count], 0
        inc byte [track]
        add word [dest], 1000h
        dec byte [reads]
        jnz getkernel

flopyoff:
        mov dx, 3f2h                    ; floppy motor off
        xor ax, ax
        out dx, al

; ******************************************************************************

; Set 800x600 Video. First try for 16 bit color. If that fails then try for
; 15 bit color.

highcolor16:
        mov cx, 114h     ; 114
        call vidtry
        jz highcolor15
        call vidcheck
        jz highcolor15
        mov bx, 04114h   ; 114
        call vidset
        jnz vidok

highcolor15:
        mov cx, 113h      ; 113
        call vidtry
        jz vidbad
        call vidcheck
        jz vidbad
        mov bx, 04113h    ; 113
        call vidset
        jnz vidok

vidbad:
        mov si, vesa
        call text
        jmp $

vidok:  mov ax, 4f07h                   ; Set first scanline and pixel = 0
        xor bx, bx
        xor cx, cx
        xor dx, dx
        int 10h
        cmp ax, 0
        jz vidbad

; ******************************************************************************

; Transfer the Protected Mode entry code, Global and Interupt Desciptor tables
; down to 0000x0000. This means all addresses system wide will match their
; physical addresses.

transfer:
        cli
        cld
        mov ax, 0
        mov es, ax
        mov di, 0
        mov ax, 1000h
        mov ds, ax
        mov si, 200h
        mov cx, 400h
        repnz movsw                     ; ds:si to es:di

; Align Enth Kernel to 1000x0000.

        mov ax, 1000h
        mov es, ax
        mov di, 0h
        mov ds, ax
        mov si, 600h
        mov cx, 7a00h
        repnz movsw                     ; ds:si to es:di

; Now we must transfer program execution to the copy of the system at 0000x0000.

        mov ax, 0                       ; set registers to code position
        mov ds, ax
        mov es, ax
        push ax                         ; segment address onto stack
        mov ax, 0h
        push ax                         ; offset address onto stack
        retf                            ; pop offset to ip, pop segment to cs

; ******************************************************************************

text:                                   ; display a message, teletype style
        lodsb                           ; load byte at ds:si into al
        or al, al                       ; 0=
        jz .end                         ; if ret then
        mov ah, 0Eh                     ; ah=0eh write char to screen teletype
        int 10h
        jmp text
.end:   ret

; ******************************************************************************

; Video Stuff

vidtry:
        mov ax, 3000h                   ; 30000h = VESA mode descriptor
        mov es, ax
        mov di, 0h
        mov ax, 04f01h                  ; VBE function 1
        int 10h
        cmp ax, 0
        ret

vidcheck:
        mov ax, 3000h
        mov es, ax
        mov di, 0h
        mov ax, [es:di]
        test ax, 1h                     ; Video Mode available
        jz .nope
        test ax, 80h                    ; Linear Frame Buffer available
.nope:  ret

vidset:
        mov ax, 04f02h
        int 10h
        cmp ax, 0
        ret

; ******************************************************************************

; Floppy read code. Not pretty.

readcheck:
        mov al, [error]
        cmp al, 0
        je .ok
        mov si, query
        call text
        jmp $
.ok:    ret

readimage:
        mov byte [retry], 0
        dec byte [head]                 ; head 0
        call readdata
        mov al, [count]
        inc al
        cmp al, byte [limit]
        jne .next
        ret
.next:  inc byte [head]                 ; head 1
        call readdata
        inc byte [track]
        add byte [count], 2
        mov al, [count]
        cmp al, byte [limit]
        jne readimage
        ret

readdata:
        xor ax, ax                      ; zero ax as [count] is 8 bit
        mov al, [count]                 ; read count
        add al, [head]                  ; + 1 if we are on the second head
        mov bx, [offset]
        mul bx
        mov bx, ax                      ; bx = offset
        mov dl, [boot]                  ; read from boot drive
        mov dh, [head]                  ; head number
        mov ah, 02h                     ; ah=02h read sectors into memory
        mov al, 18                      ; number of sectors to read
        mov cl, 01h                     ; start sector number
        mov ch, [track]                 ; track/cylinder number for floppy/hard
        int 13h                         ; do it
        jnc .done                       ; retry upon error
        mov al, [retry]                 ; retry read 3 times
        cmp al, 2                       ; third time?
        je .end
        inc byte [retry]
        mov si, query                   ; display query
        call text
        jmp readdata                    ; retry
.end    dec byte [error]                ; set error flag
        ret
.done:  mov si, dot                     ; display a dot
        call text
        ret

; ******************************************************************************

        times 510-($-$$) db 0           ; append 0xAA and 0x55 - needed for boot
        dw 0AA55h                       ;  file
